/* -*-c++-*-
 * 
 *    Source     $RCSfile: sb_test.cc,v $ 
 *    Version    $Revision: 1.1.1.1.2.1 $ 
 *    Date       $Date: 2000/02/28 13:55:41 $ 
 *    Author     Christoph Helmberg 
 * 
 --------------------------------------------------------------------------- */

// ****************************************************************************
//
//  This file is part of the class library "SBmethod", an implementation 
//  of the spectral bundle method for eigenvalue optimization.
//
//  Copyright (C) 2000  Christoph Helmberg
//
//     Konrad-Zuse-Zentrum fuer Informationstechnik Berlin
//     Takustrasse 7, D-14195 Berlin, Germany
//      
//     helmberg@zib.de     http://www.zib.de/helmberg 
// 
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// ****************************************************************************

#ident "$Id: sb_test.cc,v 1.1.1.1.2.1 2000/02/28 13:55:41 bzfhelmb Exp $"

#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <fstream.h>
#include <string.h>
#include "problem1.h"
#include "solver.h"
#include "sbparams.h"

SBparams* paramsp;
SpectralBundle* bundlep;
BaseSolver* bsolver;
Problem* problemp;
Clock* clp;

//*************************************************************************
//                             save_and_exit
//*************************************************************************

void save_and_exit(const SpectralBundle* sbp)
 {
     ofstream out;
     if (paramsp->savefile){
         out.open(paramsp->savefile);
     }
     else {
         out.open("sb_resume.dat");
     }
     out<<(sbp->get_clock()).time()<<"\n";
     paramsp->save(out);
     mat_randgen.save(out);
     sbp->get_problem()->save(out);
     sbp->get_choose_tau()->save(out);
     sbp->get_update_bundle()->save(out);
     sbp->get_terminator()->save(out);
     sbp->save(out);
     bsolver->save(out);
     out.close();
     cout<<"Terminated to resume later, intermediate output follows:"<<endl;
     int retval=paramsp->output(problemp,sbp,clp);
     exit(retval);
 }

//*************************************************************************
//                             summary_and_exit
//*************************************************************************

void summary_and_exit(const SpectralBundle* sbp)
 {
   cout<<"Terminated prematurely by User signal 1!"<<endl;
   int retval;
   retval=paramsp->output(problemp,sbp,clp);
   exit(retval);
 }

//*************************************************************************
//                             summary_and_continue
//*************************************************************************

void summary_and_continue(const SpectralBundle* sbp)
 {
   cout<<"Intermediate output caused by User signal 2:"<<endl;
   int retval;
   retval=paramsp->output(problemp,sbp,clp);
   sbp->get_terminator()->set_save_function(0);
 }

//*************************************************************************
//                             catch_signal
//*************************************************************************


void catch_signal(int sig){
 if (sig==SIGTERM){
     bundlep->get_terminator()->set_save_function(save_and_exit);
 }
 else if (sig==SIGUSR1) {
     bundlep->get_terminator()->set_save_function(summary_and_exit);
 }
 else if (sig==SIGUSR2) {
     bundlep->get_terminator()->set_save_function(summary_and_continue);
 }
 else if (sig==SIGFPE) {
   cout<<"\n\n**** ERROR: ARITHMETIC EXCEPTION, continuing anyways\n"<<endl;
 }
}


//*************************************************************************
//                               main
//*************************************************************************

int main(int argc,char **argv)
{
 SBparams params;
 paramsp=&params;

 Problem1 problem;
 problemp=&problem;
 SpectralBundle bundle;
 bundlep=&bundle;
 Clock myclock;
 clp=&myclock;
 Solver solver;
 bsolver=&solver;

 int retval;
 if ((retval=params.input(argc,argv,bsolver))) return retval; 

 Microseconds offset(0,0);
 
 ifstream fresin;
 if (params.resume){
     fresin.open(params.resume);
     if (!fresin) {
         cout<<"*** ERROR: could not open resume file "<<params.resume<<endl;
         return 1;
     }
     if (!(fresin>>offset)) {
         cout<<"*** ERROR: failed in reading time offset from resume file "<<params.resume<<endl;
         return 1;
     }
     if (!params.restore(fresin)){
         cout<<"*** ERROR: failed in reading parameters from  resume file "<<params.resume<<endl;
         return 1;
     }
 }

 if ((retval=params.initialize(&problem,&bundle,&myclock,&solver))) return retval;

 if (params.resume){
     if (!fresin){
         cerr<<"*** ERROR: main(): resume file not good before restoring mat_randgen!"<<endl;
         return 1;
     }
     mat_randgen.restore(fresin);
     if (!fresin){
         cerr<<"*** ERROR: main(): resume file not good before restoring problem!"<<endl;
         return 1;
     }
     problem.restore(fresin);
     if (!fresin){
         cerr<<"*** ERROR: main(): resume file not good before restoring choose_tau!"<<endl;
         return 1;
     }
     bundle.get_choose_tau()->restore(fresin);
     if (!fresin){
         cerr<<"*** ERROR: main(): resume file not good before restoring update_bundle!"<<endl;
         return 1;
     }
     bundle.get_update_bundle()->restore(fresin);
     if (!fresin){
         cerr<<"*** ERROR: main(): resume file not good before restoring terminator!"<<endl;
         return 1;
     }
     bundle.get_terminator()->restore(fresin);
     if (!fresin){
         cerr<<"*** ERROR: main(): resume file not good before restoring bundle!"<<endl;
         return 1;
     }
     bundle.restore(fresin);
     if (!fresin){
         cerr<<"*** ERROR: main(): resume file not good before restoring solver!"<<endl;
         return 1;
     }
     solver.restore(fresin);
     (fresin).close();
     myclock.start();
     myclock.set_offset(offset);
 }

 signal(SIGTERM,catch_signal);
 signal(SIGUSR1,catch_signal);
 signal(SIGUSR2,catch_signal);
 signal(SIGFPE,catch_signal);
 
 solver.set_out(&cout);
 if ((retval=solver.solve(&problem,&bundle,&myclock,&params,(params.resume!=0)))) return retval;

 if ((retval=params.output(&problem,&bundle,&myclock))) return retval;

 return 0;
}

/* ---------------------------------------------------------------------------
 *    Change log $Log: sb_test.cc,v $
 *    Change log Revision 1.1.1.1.2.1  2000/02/28 13:55:41  bzfhelmb
 *    Change log Reording inline functions and correction of minor details
 *    Change log
 *    Change log Revision 1.1.1.1  1999/12/15 13:47:51  bzfhelmb
 *    Change log Imported sources
 *    Change log
 *
 *    End of $Source: /kombadon/cvs/cvsroot/bzfhelmb/SBmethod/spectral/sb_test.cc,v $
 --------------------------------------------------------------------------- */
