/* -*-c++-*-
 * 
 *    Source     $RCSfile: heapsort.h,v $ 
 *    Version    $Revision: 1.1.1.1 $ 
 *    Date       $Date: 1999/12/15 13:47:48 $ 
 *    Author     Christoph Helmberg 
 * 
 --------------------------------------------------------------------------- */

#ifndef __HEAPSORT_H__
#define __HEAPSORT_H__

/* given:                                                      */
/*    n     ... number of indices to be sorted                 */
/*    ind   ... class indexable by integers,                   */
/*              a function swap(ind[i],ind[j]) is required     */
/*              which swaps the contents of ind[i] and ind[j]  */
/*    val   ... class indexable by elements of ind,            */
/*              for two vals val[i] and val[j] operator <      */
/*              (val[i]<val[j]) yielding 0/1 must be defined   */
/*                                                             */
/* Sorts the first n entries in the index array such that      */
/*                                                             */
/*        val[ind[i]]<=val[ind[j]] for 0<=i<j<n                */
/*                                                             */

template<class I,class V> void heapify(int i,int n,I& ind,const V& val);
     /* Heapify element i in index array of size n. */
     /* after this operation d[ind[i]] is a maximal */
     /* element of the values in the subtree with   */
     /* root i.                                     */

template<class I,class V> void build_heap(int n,I& ind,const V& val);
     /* Construct heap from index array of size n.  */
     /* After this operation d[ind[0]] is a maximal */
     /* element of the first n indexed values.      */

template<class I,class V> void heapsort(int n,I& ind,const V& val);
     /* sort index array */



//--------------------------------------------------------------
//                    heapsort -- implementation
//--------------------------------------------------------------

#include "mymath.h"
#include "matop.h"

template<class I,class V> inline void heapify(int i,int n,I& ind,const V& val)
{
 int k,j;
 k=2*(i+1);
 while(k<n){
   if (val[ind[k]]<val[ind[k-1]]) j=k-1;
   else j=k;
   swap(ind[k/2-1],ind[j]);
   k=2*(j+1);
 }
 if (k==n){
   if (val[ind[k/2-1]]<val[ind[k-1]]){
     swap(ind[k/2-1],ind[k-1]);
     return;
   }
 }
 k=k/2-1;
 j=(k+1)/2-1;
 while((j>=i)&&(val[ind[j]]<val[ind[k]])){
   swap(ind[j],ind[k]);
   k=j;
   j=(k+1)/2-1;
 }
}

template<class I,class V> inline void build_heap(int n,I& ind,const V& val)
{
 for(int i=n/2;--i>=0;) heapify(i,n,ind,val);
}

template<class I,class V> inline void heapsort(int n,I& ind,const V& val)
{
 int i;

 //build heap
 build_heap(n,ind,val);

 //extract greatest and rebuild heap
 for(i=n;--i>=1;){
     swap(ind[i],ind[0]);
     heapify(0,i,ind,val);
 }
}

template<class V> inline void heapify_Integer(int i,int n,Integer* ind,const V& val)
{
 int k,j;
 Integer h=ind[i];
 k=2*(i+1);
 while(k<n){
   if (val[ind[k]]<val[ind[k-1]]) j=k-1;
   else j=k;
   ind[k/2-1]=ind[j];
   k=2*(j+1);
 }
 if (k==n){
   if (val[h]<val[ind[k-1]]){
     ind[k/2-1]=ind[k-1];
     ind[k-1]=h;
     return;
   }
 }
 k=k/2-1;
 j=(k+1)/2-1;
 while((j>=i)&&(val[ind[j]]<val[h])){
   ind[k]=ind[j];
   k=j;
   j=(k+1)/2-1;
 }
 ind[k]=h;
}

template<class V> inline void build_heap_Integer(int n,Integer* ind,const V& val)
{
 for(int i=n/2;--i>=0;) heapify_Integer(i,n,ind,val);
}

template<class V> inline void heapsort_Integer(int n,Integer* ind,const V& val)
{
 int i;

 //build heap
 build_heap_Integer(n,ind,val);

 //extract greatest and rebuild heap
 for(i=n;--i>=1;){
     swap(ind[i],ind[0]);
     heapify_Integer(0,i,ind,val);
 }
}

#endif

/* ---------------------------------------------------------------------------
 *    Change log $Log: heapsort.h,v $
 *    Change log Revision 1.1.1.1  1999/12/15 13:47:48  bzfhelmb
 *    Change log Imported sources
 *    Change log
 *
 *    End of $Source: /kombadon/cvs/cvsroot/bzfhelmb/SBmethod/tools/heapsort.h,v $
 --------------------------------------------------------------------------- */
